//
// CSE223B Lab 3: Distributed, Multi-party Chat Client
//
// Copyright(c) 2003-6 Alex C. Snoeren <snoeren@cs.ucsd.edu>
//
// You should not modify this code as part of your assignment.
//

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/wait.h>
#include <netdb.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <iostream>
#include <readline/readline.h>

using namespace std;

#define MAXCHATLINE 255

int fd;
char ctrlu = '\15';
bool interactive = true;
bool done = false;
fd_set rfds;

void
allowContextSwitch(void)
{
  int pid, status;

  pid = fork();
  if(pid == 0){
    exit(0);
  } else if(pid > 0){
    waitpid(pid, &status, 0);
  }
}

void
sendMessage(char *buff)
{
  cout << ctrlu;
  if(!buff) {
    cout << "Disconnecting..." << endl;
    if(interactive)
      rl_callback_handler_remove();
    done = true;
    shutdown(fileno(stdin), SHUT_RD);
    shutdown(fd, SHUT_WR);
  } else {
    // Ignore blank lines
    if(buff[0] != '\0') {
      if(write(fd, buff, strlen(buff)+1) != strlen(buff)+1) {
	perror("write");
	exit(1);
      }
      // Make sure we allow for messages to pass in batch mode
      if(!interactive)
	allowContextSwitch();
    }
  }
}

// This is a hack--we assume we never receive less than a line at a time.
int
readln(int fd, char *buf, int len)
{
  int n, i = 0;
  char c;
  while(i < len) {
    if((n = read(fd, &buf[i], 1)) < 0) {
      perror("readln");
      exit(1);
    }
    if(n == 0)
      return 0;
    // Don't print new lines
    if(buf[i] == '\n') {
      buf[i] = '\0';
      break;
    }
    if(buf[i] == '\0')
      break;
    ++i;
  }
  return i;
}

void
do_loop()
{
  int n;
  char buff[MAXCHATLINE];

  while(1) {
    FD_ZERO(&rfds);
    FD_SET(fd, &rfds);
    if(!done)
    FD_SET(fileno(stdin), &rfds);
    
    // Block until there's an event to handle
    if(select(FD_SETSIZE, &rfds, NULL, NULL, 0) < 0) {
      perror("select");
      exit(1);
    }
    
    if(FD_ISSET(fileno(stdin), &rfds)) {
      if(interactive)
	rl_callback_read_char();
      else {
	char buff[MAXCHATLINE];
	if(fgets(buff, MAXCHATLINE, stdin)) {
	  // Trim newline
	  buff[strlen(buff)-1] = '\0';
	  sendMessage(buff);
	}
	else
	  sendMessage(NULL);
      }
    }
    
    if(FD_ISSET(fd, &rfds)) {
      if((n = readln(fd, buff, MAXCHATLINE)) == 0) {
	cout << ctrlu << "Server disconnected." << endl;
	exit(0);
      }
      // Null-terminate string before writing
      buff[n] = '\0';
      if(interactive)
	rl_save_prompt();
      cout << ctrlu << buff << endl;
      if(interactive) {
	rl_on_new_line();
	rl_restore_prompt();
	rl_redisplay();
      }
    }
  }
}


main(int argc, char*argv[])
{
  int n = 0;

  if((argc < 4) || (argc > 5)) {
    cerr << "Usage: " << argv[0] << " [-i] host port handle" << endl;
    exit(1);
  }

  if(strcmp(argv[1], "-i") == 0) {
    interactive = false;
    ++n;
    setlinebuf(stdin);
    cout << "Batch mode" << endl;
  }

  struct hostent *hp;
  if(!(hp = gethostbyname(argv[++n]))) {
    cerr << argv[1] << ": Unknown host" << endl;
    exit(1);
  }

  // new socket
  fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if(fd < 0) {
    perror("socket");
    exit(1);
  }

  // connect to destination
  sockaddr_in sin;
  bzero(&sin, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = htons(atoi(argv[++n]));
  sin.sin_addr.s_addr = *(in_addr_t *)hp->h_addr;
  if(connect(fd, (sockaddr*) &sin, sizeof(sin)) < 0) {
    perror("connect");
    exit(1);
  }
  ++n;
  if(write(fd, argv[n], strlen(argv[n])+1) != strlen(argv[n])+1) {
    perror("write");
    exit(1);
  }  

  setlinebuf(stdout);
  if(interactive)
    rl_callback_handler_install("> ", sendMessage);
  do_loop();
}
